{-# LANGUAGE GADTs, FlexibleInstances #-}
module ParserBreadthFirst where

import Control.Monad (MonadPlus(..), liftM2)
import Data.Char (isDigit, ord)

import Program

    -- three primitive instructions
data ParserInstruction a where
    Symbol :: ParserInstruction Char
    MZero  :: ParserInstruction a
    MPlus  :: Parser a -> Parser a -> ParserInstruction a

type Parser a = Program ParserInstruction a

symbol = singleton Symbol
instance MonadPlus (Program ParserInstruction) where
    mzero       = singleton MZero
    p `mplus` q = singleton $ MPlus p q


    -- interpreter, defined by pattern matching
    -- breadth-first
interpret :: Parser a -> String -> [a]
interpret p = interpret' (expand p)
    where
    interpret' :: [Parser a] -> String -> [a]
    interpret' ps []     = [a | Return a <- ps]
    interpret' ps (c:cs) =
        interpret' [p | (Symbol `Then` is) <- ps, p <- expand (is c)] cs
    
    expand :: Parser a -> [Parser a]
    expand (MPlus p q `Then` is) = expand (p >>= is) ++ expand (q >>= is)
    expand (MZero     `Then` is) = []
    expand x                     = [x]


    -- example gramamr, try for instance
    --    interpret number "42"
    --    interpret (many number) "3142"
satisfies p = symbol >>= \c -> if p c then return c else mzero 
many  p     = return [] `mplus` many1 p
many1 p     = liftM2 (:) p (many p) 
digit       = satisfies isDigit >>= \c -> return (ord c - ord '0')
number      = many1 digit >>= return . foldl (\x d -> 10*x + d) 0
