{-# LANGUAGE GADTs, FlexibleInstances #-}
module ParserDepthFirst where

import Control.Monad (MonadPlus(..), liftM2)
import Data.Char (isDigit, ord)

import Program

    -- three primitive instructions
data ParserInstruction a where
    Symbol :: ParserInstruction Char
    MZero  :: ParserInstruction a
    MPlus  :: Parser a -> Parser a -> ParserInstruction a

type Parser a = Program ParserInstruction a

symbol = singleton Symbol
instance MonadPlus (Program ParserInstruction) where
    mzero       = singleton MZero
    p `mplus` q = singleton $ MPlus p q


    -- interpreter, defined by pattern matching
    -- depth-first
interpret :: Parser a -> String -> [a]
interpret (Return a)            s = if null s then [a] else []
interpret (Symbol    `Then` is) s = case s of
    c:cs -> interpret (is c) cs
    []   -> []
interpret (MZero     `Then` is) s = []
interpret (MPlus p q `Then` is) s =
    interpret (p >>= is) s ++ interpret (q >>= is) s


    -- example grammar, try for instance
    --     > interpret number "42"
    --     > interpret (many number) "3142"
satisfies p = symbol >>= \c -> if p c then return c else mzero 
many  p     = return [] `mplus` many1 p
many1 p     = liftM2 (:) p (many p) 
digit       = satisfies isDigit >>= \c -> return (ord c - ord '0')
number      = many1 digit >>= return . foldl (\x d -> 10*x + d) 0
