{-# LANGUAGE GADTs, FlexibleInstances #-}
module StackMachineMonad where

import Control.Monad ((>>))
import Program

-- Representation
data StackInstruction a where
    Pop  :: StackInstruction Int
    Push :: Int -> StackInstruction ()

type StackProgram a = Program StackInstruction a

-- Interpreter
type Stack a = [a]

interpret :: StackProgram a -> (Stack Int -> a)
interpret (Push a `Then` is) stack     = interpret (is ()) (a:stack)
interpret (Pop    `Then` is) (a:stack) = interpret (is a ) stack
interpret (Return a)         stack     = a

-- Concatenation and interface
pop  :: StackProgram Int
push :: Int -> StackProgram ()
pop  = singleton Pop
push = singleton . Push

-- Examples, try for instance
    --      > interpret example3 [7,11] 
    --      > interpret example4 [7,11,13]
    
    -- examples using instructions directly
example   = Push 5  `Then` \() ->
            Push 42 `Then` \() ->
            Pop `Then` Return
example2  = Pop `Then` \a ->
            Pop `Then` \b ->
            Push (a+b) `Then`
            Return
example3  = Pop `Then` \a -> Pop `Then` \b -> Return (a*b)

    -- examples by concatenating programs
    -- 
    -- These examples use a new combinator  (>>)  which is
    -- defined in  Control.Monad  as
    --   m >> n  =  m >>= \_ -> n
    -- In other words, it's a variant of  (>>=)  which simply
    -- ignores the result of the first argument  m .
exampleTwice  = example >> example
replace a     = pop >> push a
example4      = example3 >>= \ab ->
                pop      >>= \c  ->
                return (ab*c)

popN 1 = pop
popN n = pop >> popN (n-1)
